/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#ifndef __GAME_H
#define __GAME_H

#include <stdio.h>

#include <set>
#include <list>
#include <vector>
#include <string>

typedef list<string> path_container;

enum merge_t {
	merge_no, // All required roms in every set
	merge_differential, // Only the unique roms in every set
	merge_parent, // All roms in the parent set
	merge_any, // Any merge allowed
	merge_disable // Check disabled
};

// ------------------------------------------------------------------------
// Group

#define GROUP_UNDEFINED "<undefined>"

typedef list<string> group_container;

#define TYPE_UNDEFINED "<undefined>"

typedef list<string> type_container;

// ------------------------------------------------------------------------
// Game

class game;

typedef list<const game*> pgame_container;
typedef vector<const game*> pgame_array;

class game {
	static const unsigned flag_working = 0x0001;
	static const unsigned flag_resource = 0x0002;
	static const unsigned flag_software = 0x0004;
	static const unsigned flag_vector = 0x0008;
	static const unsigned flag_sound = 0x0010;
	static const unsigned flag_color = 0x0020;
	static const unsigned flag_alias = 0x0040;

	unsigned flag;

	void flag_set(bool value, unsigned mask) {
		if (value)
			flag |= mask;
		else
			flag &=~mask;
	}

	bool flag_get(unsigned mask) const {
		return (flag & mask) != 0;
	}

	// game information
	string name;
	string romof;
	string cloneof;
	string sampleof;
	string description;
	string year;
	string manufacturer;
	mutable string software_path;
	unsigned sizex;
	unsigned sizey;

	mutable string group;
	mutable string type;
	mutable unsigned time;

	mutable unsigned size; // cached size in bytes of the roms

	mutable path_container rzs; // set of rom zip for the game

	mutable pgame_container clone_bag; // clones
	mutable const game* parent; // parent

public:
	game();
	game(const string& Aname);
	game(const game&);
	~game();

	void name_set(const string& A) { name = A; }
	void cloneof_set(const string& A) { cloneof = A; }
	void romof_set(const string& A) { romof = A; }
	void description_set(const string& A) { description = A; }
	void year_set(const string& A) { year = A; }
	void manufacturer_set(const string& A) { manufacturer = A; }
	void software_path_set(const string& A) const { software_path = A; }
	void working_set(bool A) { flag_set(A,flag_working); }
	void resource_set(bool A) { flag_set(A,flag_resource); }
	void software_set(bool A) { flag_set(A,flag_software); }
	void time_set(unsigned A) const { time = A; }
	void group_set(const string& A) const { group = A; }
	void type_set(const string& A) const { type = A; }
	void vector_set(bool A) { flag_set(A,flag_vector); }
	void size_set(unsigned Asize) const { size = Asize; }
	void sound_set(bool A) { flag_set(A,flag_sound); }
	void color_set(bool A) { flag_set(A,flag_color); }
	void alias_set(bool A) { flag_set(A,flag_alias); }
	void sizex_set(unsigned A) { sizex = A; }
	void sizey_set(unsigned A) { sizey = A; }
	void parent_set(const game* A) const { parent = A; }

	void rom_zip_set_add(const string& Afile) const;

	const string& name_get() const { return name; }
	const string& cloneof_get() const { return cloneof; }
	const string& romof_get() const { return romof; }
	const string& description_get() const { return description; }
	const string& year_get() const { return year; }
	const string& manufacturer_get() const { return manufacturer; }
	const string& software_path_get() const { return software_path; }
	bool working_get() const { return flag_get(flag_working); }
	bool resource_get() const { return flag_get(flag_resource); }
	bool software_get() const { return flag_get(flag_software); }
	unsigned time_get() const { return time; }
	string group_get() const { return group; }
	string type_get() const { return type; }
	bool vector_get() const { return flag_get(flag_vector); }
	unsigned size_get() const { return size; }
	unsigned clone_get() const { return clone_bag.size(); }
	bool sound_get() const { return flag_get(flag_sound); }
	bool color_get() const { return flag_get(flag_color); }
	bool alias_get() const { return flag_get(flag_alias); }
	unsigned sizex_get() const { return sizex; }
	unsigned sizey_get() const { return sizey; }

	string group_derived_get() const;
	string type_derived_get() const;

	pgame_container& clone_bag_get() const { return clone_bag; }
	void clone_bag_erase() const { clone_bag.clear(); }
	const game* parent_get() const { return parent; }
	const game* bios_get() const;

	const path_container& rom_zip_set_get() const { return rzs; }

	bool is_present() const {
		return size_get() == 0 || rom_zip_set_get().size() > 0;
	}

	bool is_good() const {
		return working_get() && color_get() && sound_get();
	}

	const game* clone_best_get() const;
};

typedef list<game> game_container;

struct game_by_name_less : binary_function<game,game,bool> {
	bool operator()(const game& A, const game& B) const {
		return A.name_get().compare(B.name_get()) < 0;
	}
};

struct game_by_play_less : binary_function<game,game,bool> {
	bool operator()(const game& A, const game& B) const {
		if (!A.working_get() && B.working_get()) return true;
		if (A.working_get() && !B.working_get()) return false;
		if (!A.color_get() && B.color_get()) return true;
		if (A.color_get() && !B.color_get()) return false;
		if (!A.sound_get() && B.sound_get()) return true;
		/* if (A.sound_get() && !B.sound_get()) return false; */
		return false;
	}
};

typedef set<game,game_by_name_less> game_by_name_set;

class game_set : public game_by_name_set {
public:
	typedef game_by_name_set::const_iterator const_iterator;
	typedef game_by_name_set::iterator iterator;

	bool test() const;
	bool load(FILE* f);
	void sync();

	bool is_tree_rom_of_present(const string& name, merge_t type) const;
	bool is_game_neogeo(const string& name) const;
	game_set::const_iterator root_rom_of_get(const string& name) const;
	game_set::const_iterator root_clone_of_get(const string& name) const;
	bool is_game_rom_of(const string& name_son, const string& name_parent) const;
	bool is_game_clone_of(const string& name_son, const string& name_parent) const;
};

inline bool pgame_combine_less(const game* A, const game* B, bool (*FA)(const game*, const game*), bool (*FB)(const game*, const game*) ) {
	if (FA(A,B)) return true;
	if (FA(B,A)) return false;
	return FB(A,B);
}

inline bool pgame_combine_less(const game* A, const game* B, bool (*FA)(const game*, const game*), bool (*FB)(const game*, const game*), bool (*FC)(const game*, const game*) ) {
	if (FA(A,B)) return true;
	if (FA(B,A)) return false;
	if (FB(A,B)) return true;
	if (FB(B,A)) return false;
	return FC(A,B);
}

inline bool pgame_by_desc_less(const game* A, const game* B) {
	return A->description_get() < B->description_get();
}

inline bool pgame_by_rootdesc_less(const game* A, const game* B) {
	while (A->parent_get()) A = A->parent_get();
	while (B->parent_get()) B = B->parent_get();
	return pgame_by_desc_less(A,B);
}

inline bool pgame_by_leveldesc_less(const game* A, const game* B) {
	const unsigned stack_max = 16;
	const game* Astack_begin[stack_max];
	const game** Astack_end = Astack_begin;
	const game* Bstack_begin[stack_max];
	const game** Bstack_end = Bstack_begin;
	while (A) {
		*Astack_end++ = A;
		A = A->parent_get();
	}
	while (B) {
		*Bstack_end++ = B;
		B = B->parent_get();
	}
	while (Astack_begin != Astack_end && Bstack_begin != Bstack_end) {
		--Astack_end;
		--Bstack_end;
		if ((*Astack_end)->software_get() && !(*Bstack_end)->software_get())
			return true;
		if (!(*Astack_end)->software_get() && (*Bstack_end)->software_get())
			return false;
		if ((*Astack_end)->description_get() != (*Bstack_end)->description_get())
			return pgame_by_desc_less(*Astack_end,*Bstack_end);
	}
	return Astack_begin == Astack_end && Bstack_begin != Bstack_end;
}

inline bool pgame_by_manufacturer_less(const game* A, const game* B) {
	return A->manufacturer_get() < B->manufacturer_get();
}

inline bool pgame_by_year_less(const game* A, const game* B) {
	return A->year_get() < B->year_get();
}

inline bool pgame_by_time_less(const game* A, const game* B) {
	return A->time_get() > B->time_get();
}

inline bool pgame_by_group_less(const game* A, const game* B) {
	return A->group_derived_get() < B->group_derived_get();
}

inline bool pgame_by_type_less(const game* A, const game* B) {
	return A->type_derived_get() < B->type_derived_get();
}

inline bool pgame_by_size_less(const game* A, const game* B) {
	return A->size_get() > B->size_get();
}

inline bool pgame_by_clone_less(const game* A, const game* B) {
	return (A->parent_get()!=0) < (B->parent_get()!=0);
}

inline bool pgame_by_name_less(const game* A, const game* B) {
	return A->name_get() < B->name_get();
}

#endif
