/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "game.h"
#include "menu.h"
#include "text.h"
#include "bitmap.h"
#include "info.h"
#include "crc.h"
#include "unzip.h"

#include <pc.h>
#include <allegro.h>
#include <stdlib.h>
#include <dos.h>
#include <dirent.h>
#include <stdio.h>
#include <time.h>
#include <unistd.h>
#include <sys/exceptn.h>
#include <sys/stat.h>

#include <strstream>
#include <fstream>

#define FILE_FONT "mm.fnt"
#define FILE_INFO "info.txt"
#define FILE_CFG "mm.cfg"

#define FILE_MAME "mame.exe"
#define FILE_CFG_MAME "mame.cfg"
#define FILE_MESS "mess.exe"
#define FILE_CFG_MESS "mess.cfg"

// ------------------------------------------------------------------------
// Generic

extern "C" void video_log_va(const char* format, va_list arg) {
	(void)format;
	(void)arg;
}

/*
INT 2F - SMARTDRV v4.00+ - FLUSH BUFFERS (COMMIT CACHE)
	AX = 4A10h
	BX = 0001h
Note:   this function is also supported by PC-Cache v8.0.
*/
static void flush_cache() {
	union REGS r;
	r.x.ax = 0x4A10;
	r.x.bx = 0x0001;
	int86(0x2F, &r, &r);
}

void sound_signal(void) {
	unsigned i;
	for(i=0;i<10;++i) {
		sound(800);
		delay(10);
		nosound();
		delay(10);
	}
}

string basename(const string& s) {
	int i = s.find('.');
	if (i==string::npos)
		return s;
	else
		return string(s,0,i-1);
}

string slash_add(const string& s) {
	string r = s;
	if (r.length() && r[r.length()-1]!='\\' && r[r.length()-1]!='/')
		r += '/';
	return r;
}

bool file_crc(const string& file, unsigned& crc) {
	struct stat st;

	if (stat(file.c_str(), &st)!=0)
		return false;

	FILE* f = fopen(file.c_str(),"rb");
	if (!f)
		return false;

	unsigned char* data = (unsigned char*)operator new(st.st_size);

	if (fread(data,st.st_size,1,f)!=1) {
		fclose(f);
		return false;
	}

	fclose(f);

	crc = crc32(0,data,st.st_size);

	operator delete(data);
	return true;
}

int accesszip(const char* zip_file, const char* ent_file, unsigned* crc) {
	ZIP* zip;
	struct zipent* ent;

	if (access(zip_file,F_OK)!=0)
		return -1;

	zip = openzip(zip_file);
	if (!zip)
		return -1;

	while ((ent = readzip(zip))!=0) {
		if (strcmp(ent->name,ent_file)==0) {
			*crc = ent->crc32;
			closezip(zip);
			return 0;
		}
	}
	closezip(zip);
	return -1;
}

// ------------------------------------------------------------------------
// Conversion

string tristate2string(tristate_t v) {
	switch (v) {
		case include : return "include";
		case exclude : return "exclude";
		case exclude_not : return "exclude_not";
	}
	return "";
}

tristate_t string2tristate(const string& v) {
	if (v=="include") {
		return include;
	} else if (v=="exclude") {
		return exclude;
	} else if (v=="exclude_not") {
		return exclude_not;
	}
	return (tristate_t)-1;
}

string sort2string(game_sort_t v) {
	switch (v) {
		case sort_by_group : return "group";
		case sort_by_root_name : return "parent";
		case sort_by_name : return "name";
		case sort_by_time : return "time";
		case sort_by_year : return "year";
		case sort_by_manufacturer : return "manufacturer";
		case sort_by_type : return "type";
		case sort_by_size : return "size";
	}
	return "";
}

game_sort_t string2sort(const string& v) {
	if (v=="group") {
		return sort_by_group;
	} else if (v=="name") {
		return sort_by_name;
	} else if (v=="parent") {
		return sort_by_root_name;
	} else if (v=="time") {
		return sort_by_time;
	} else if (v=="year") {
		return sort_by_year;
	} else if (v=="manufacturer") {
		return sort_by_manufacturer;
	} else if (v=="type") {
		return sort_by_type;
	} else if (v=="size") {
		return sort_by_size;
	}
	return (game_sort_t)-1;
}

// ------------------------------------------------------------------------                                                                        // ------------------------------------------------------------------------
// Load

void scan_game(const game_set& gar, const string& path, const string& name) {
	game_set::const_iterator i = gar.find( game(name) );
	if (i!=gar.end()) {
		i->rom_zip_set_add(path);
	}
}

void scan_dir(const game_set& gar, const string& dir) {
	DIR* d = opendir(dir.c_str());
	if (!d) {
		cerr << "warning: error opening roms directory " << dir << endl;
		return;
	}

	struct dirent* dd;

	while ((dd = readdir(d))!=0) {
		if (strstr(dd->d_name,".ZIP")!=0 || strstr(dd->d_name,".zip")!=0) {
			const char* file = dd->d_name;
			const char* dot = strchr(dd->d_name,'.');
			string path = string(dir) + "/" + string(dd->d_name);
			string name = string(file,dot-file);
			scan_game(gar,path,name);
		}
	}

	closedir(d);
}

void scan_dirlist(const game_set& gar, const string& dirlist) {
	unsigned i = 0;
	while (i<dirlist.length()) {
		unsigned end = dirlist.find(';',i);
		if (end==string::npos) {
			scan_dir(gar, string( dirlist, i ));
			i = dirlist.size();
		} else {
			scan_dir(gar, string( dirlist, i, end-i ));
			i = end + 1;
		}
	}
}

void scan_cfg(menu_state& rs, const game_set& gar, unsigned& orientation, double& gamma, double& brightness) {
	char* s;

	orientation = 0;
	s = get_config_string("config", "ror", "no");
	if (stricmp(s,"yes")==0) {
		orientation ^= ORIENTATION_FLIP_XY | ORIENTATION_MIRROR_X;
	}

	s = get_config_string("config", "rol", "no");
	if (stricmp(s,"yes")==0) {
		orientation ^= ORIENTATION_FLIP_XY | ORIENTATION_MIRROR_Y;
	}

	s = get_config_string("config", "flipx", "no");
	if (stricmp(s,"yes")==0) {
		orientation ^= ORIENTATION_MIRROR_X;
	}

	s = get_config_string("config", "flipy", "no");
	if (stricmp(s,"yes")==0) {
		orientation ^= ORIENTATION_MIRROR_Y;
	}

	gamma = get_config_float("config", "gamma", 1.0);
	brightness = get_config_float("config", "brightness", 1.0);

	switch (rs.emulator) {
		case emulator_mame : s = get_config_string("directory", "rompath", "roms"); break;
		case emulator_mess : s = get_config_string("directory", "biospath", "bios"); break;
	}
	scan_dirlist(gar, s);

	if (rs.emulator == emulator_mess) {
		s = get_config_string("directory", "softwarepath", "software");
		rs.software_path = s;

		for(game_set::const_iterator i=gar.begin();i!=gar.end();++i) {
			char* sc = get_config_string(const_cast<char*>(i->name_get().c_str()), "softwarepath", 0);
			if (sc)
				i->software_path_set(sc);
			else
				i->software_path_set(rs.software_path);
		}

	}

	s = get_config_string("directory", "snap", "snap");
	rs.snap_create_path = slash_add(s);
	rs.snap_path = rs.snap_create_path;

	s = get_config_string("directory", "flyers", "flyers");
	rs.flyers_path = s;

	s = get_config_string("directory", "altss", "");
	string altss = s;

	if (altss.length()) {
		if (rs.snap_path.length())
			rs.snap_path += ";";
		rs.snap_path += altss;
	}

	s = get_config_string("directory", "crc", "crc");
	rs.crc_path = slash_add(s);
}

void scan_info(game_set& gar, const string& info, const string& mame) {
	struct stat st_info;
	struct stat st_mame;
	int err_info;
	int err_mame;

	err_info = stat(info.c_str(), &st_info);
	err_mame = stat(mame.c_str(), &st_mame);

	if (err_mame==0 && (err_info!=0 || st_info.st_mtime < st_mame.st_mtime)) {
		cerr << "warning: creating the file " << info << endl;
		char cmd[256];
		sprintf(cmd,"%s -listinfo > %s",mame.c_str(),info.c_str());
		__djgpp_exception_toggle();
		int r = system(cmd);
		__djgpp_exception_toggle();

		if (r == -1) {
			cerr << "warning: DOS error " << -r << endl;
		} else if (r & 0xFF != 0) {
			cerr << "warning: MAME error " << (r & 0xFF) << endl;
		} else if ((r >> 8) & 0xFF != 0) {
			cerr << "warning: MAME signal " << ((r >> 8) & 0xFF) << endl;
		}
	}

	FILE* f = fopen(info.c_str(),"rt");
	if (!f) {
		cerr << "error: file " << info << " not found!" << endl;
		cerr << "error: run manually the command: MAME -listinfo > " << info << endl;
		exit(EXIT_FAILURE);
	}
	if (!gar.load(f)) {
		cerr << "error: loading game informations from file " << info << " at row " << info_row_get()+1 << " column " << info_col_get()+1 << endl;
		exit(EXIT_FAILURE);
	}
	if (!gar.test()) {
		cerr << "error: in informations from file " << info << endl;
#if 1
		cerr << "error: resume" << endl;
#else
		exit(EXIT_FAILURE);
#endif
	}
	fclose(f);
}

void scan_software_by_sys(game_container& gac, const crc_list& cl, const string& software, const string& parent)
{
	string software_dir = software + parent;
	DIR* dir = opendir(software_dir.c_str());
	if (!dir)
		return;

	struct dirent* ddir;
	while ((ddir = readdir(dir))!=0) {
		if (strstr(ddir->d_name,".ZIP")!=0 || strstr(ddir->d_name,".zip")!=0) {
			const char* file = ddir->d_name;
			const char* dot = strchr(ddir->d_name,'.');
			string name = string(file,dot-file);

			game g;
			g.name_set( parent + "/" + name );
			g.software_set(true);
			g.cloneof_set(parent);
			g.romof_set(parent);

			string path = software_dir + "/" + ddir->d_name;
			g.rom_zip_set_add(path);

			crc_info ci;
			bool found = false;
			if (cl.bag.begin() != cl.bag.end()) {
				ZIP* zip = openzip(path.c_str());
				if (zip) {
					struct zipent* ent;
					while ((ent = readzip(zip))!=0) {
						ci.crc = ent->crc32;
						crc_info_set::const_iterator i = cl.bag.find(ci);
						if (i!=cl.bag.end()) {
							g.description_set(i->description);
							g.manufacturer_set(i->manufacturer);
							g.year_set(i->year);
							found = true;
							break;
						}
					}
					closezip(zip);
				}
			}
			if (!found) {
				g.description_set(name);
			}
			gac.insert(gac.end(),g);
		}
	}

	closedir(dir);
}

void scan_software(game_container& gac, const game_set& gar, const crc_list_set& cls) {
	for(game_set::const_iterator i=gar.begin();i!=gar.end();++i) {
		int ptr = 0;
		while (ptr < i->software_path_get().length()) {
			string path = slash_add(token_get(i->software_path_get(),ptr," ;"));
			token_skip(i->software_path_get(),ptr," ;");

			crc_list cl;
			cl.name = i->name_get();
			crc_list_set::const_iterator j = cls.find(cl);
			if (j!=cls.end())
				scan_software_by_sys(gac,*j,path,i->name_get());
			else
				scan_software_by_sys(gac,cl,path,i->name_get());
		}
	}
}

bool scan_game_complete_by_crc(game& g, unsigned crc, const crc_list& cl) {
	crc_info ci;
	ci.crc = crc;
	crc_info_set::const_iterator i = cl.bag.find(ci);
	if (i!=cl.bag.end()) {
		g.description_set(i->description);
		g.manufacturer_set(i->manufacturer);
		g.year_set(i->year);
		return true;
	}
	return false;
}

bool scan_game_complete_by_file(game& g, const string& parent, const string& file, const crc_list& cl, const string& software) {
	int ptr = 0;
	while (ptr < software.length()) {
		string dir = slash_add(token_get(software,ptr," ;"));
		unsigned crc;
		string path;

		path = dir + file;
		if (access(path.c_str(),F_OK)==0) {
			if (!file_crc(path,crc))
				return false;
			return scan_game_complete_by_crc(g,crc,cl);
		}

		path = dir + parent + "/" + file;
		if (access(path.c_str(),F_OK)==0) {
			if (!file_crc(path,crc))
				return false;
			return scan_game_complete_by_crc(g,crc,cl);
		}

		path = dir + basename(file) + ".zip";
		if (accesszip(path.c_str(),file.c_str(),&crc)==0) {
			return scan_game_complete_by_crc(g,crc,cl);
		}

		path = dir + parent + "/" + basename(file) + ".zip";
		if (accesszip(path.c_str(),file.c_str(),&crc)==0) {
			return scan_game_complete_by_crc(g,crc,cl);
		}

		token_skip(software,ptr," ;");
	}
	return false;
}

bool scan_game_complete_by_alias(game& g, const string& parent, const string& s, int& ptr, const crc_list& cl, const string& software) {
	while (ptr < s.length()) {
		string file = token_get(s,ptr," ");
		if (file.length() && file[0]!='-')
			if (scan_game_complete_by_file(g,parent,file,cl,software))
				return true;
		token_skip(s,ptr," ");
	}
	return false;
}

void scan_alias(game_set& gar, game_container& gac, const string& cfg, const crc_list_set& cls) {
	ifstream f(cfg.c_str());
	if (!f) {
		cerr << "warning: error opening " << cfg << " for alias" << endl;
		return;
	}

	game_set::const_iterator parent_g = gar.end();
	crc_list_set::const_iterator parent_cl;

	while (!f.eof()) {
		string s;
		getline(f,s,'\n');

		if (s.length() && s[0]!='#') {
			int ptr = 0;

			token_skip(s,ptr," ");
			string first = token_get(s,ptr," =");

			if (first.length()>=2 && first[0]=='[' && first[first.length()-1]==']') {
				string name = string(first,1,first.length()-2);
				parent_g = gar.find( game(name) );
				if (parent_g != gar.end()) {
					crc_list cl;
					cl.name = parent_g->name_get();
					parent_cl = cls.find(cl);
				}
			} else {
				if (parent_g != gar.end()) {
					game g;
					g.name_set( parent_g->name_get() + "/" + first );
					g.software_set(true);
					g.alias_set(true);
					g.cloneof_set(parent_g->name_get());
					g.romof_set(parent_g->name_get());
					token_skip(s,ptr," =");
					if (parent_cl==cls.end() || !scan_game_complete_by_alias(g,parent_g->name_get(),s,ptr,*parent_cl,parent_g->software_path_get())) {
						g.description_set( first );
					}
					gac.insert(gac.end(),g);
				}
			}
		}
	}

	f.close();
}

void group_insert(group_container& grp, const string& name) {
	group_container::iterator j;
	for(j=grp.begin();j!=grp.end();++j) {
		if (*j == name)
			break;
	}

	if (j==grp.end())
		grp.insert(grp.end(), name);
}

void type_insert(type_container& typ, const string& name) {
	type_container::iterator j;
	for(j=typ.begin();j!=typ.end();++j) {
		if (*j == name)
			break;
	}

	if (j==typ.end())
		typ.insert(typ.end(), name);
}

bool scan_game(game_set& gar, group_container& grp, type_container& typ, const string& name, const string& group, const string& type, const string& time) {
	game_set::const_iterator i = gar.find( game( name ) );
	if (i==gar.end())
		return false;

	i->group_set(group);
	i->type_set(type);
	i->time_set(atoi(time.c_str()));

	group_insert(grp,group);
	type_insert(typ,type);

	return true;
}

void scan_grp(menu_state& rs, game_set& gar, group_container& grp, type_container& typ, const string& file) {
	ifstream f(file.c_str(), ios::in );
	if (!f) {
		cerr << "warning: creating a new configration file " << file << endl;
		group_insert(grp,"Very Good");
		group_insert(grp,"Good");
		group_insert(grp,"Bad");
		if (rs.emulator == emulator_mess) {
			type_insert(typ,"Computer");
			type_insert(typ,"Consolle");
			type_insert(typ,"Application");
		}
		type_insert(typ,"Arcade");
		type_insert(typ,"Shot'em Up");
		type_insert(typ,"Bet'em Up");
		type_insert(typ,"Fighting");
		type_insert(typ,"Gun");
		type_insert(typ,"Puzzle");
		type_insert(typ,"RPG");
		type_insert(typ,"Sport");
		type_insert(typ,"Bowling");
		type_insert(typ,"Breakout");
		type_insert(typ,"Filler");
		type_insert(typ,"Racing");
		type_insert(typ,"Flipper");
	} else {
		while (!f.eof()) {
			string s;
			getline(f,s,'\n');

			string cmd;

			unsigned i = 0;

			while (i < s.length() && isspace(s[i]))
				++i;

			if (i<s.length() && s[0]!='#') {
				while (i < s.length() && !isspace(s[i])) {
					cmd += s[i];
					++i;
				}

				while (i < s.length() && isspace(s[i]))
					++i;

				string value = string(s,i);
				while (value.length() && isspace(value[value.length()-1]))
					value.erase(value.length()-1,1);

				if (cmd == "group" || cmd == "type" || cmd == "select_group" || cmd == "select_type") {
					string name;

					while (i < s.length() && isspace(s[i]))
						++i;

					if (i >= s.length() || s[i]!='"')
						goto err;
					++i;

					while (i < s.length() && s[i]!='"') {
						name += s[i];
						++i;
					}
	
					if (name.length()==0 || i >= s.length() || s[i]!='"')
						goto err;
					++i;

					while (i < s.length() && isspace(s[i]))
						++i;

					if (i<s.length())
						goto err;

					if (cmd == "group") {
						group_insert(grp,name);
					} else if (cmd == "type") {
						type_insert(typ,name);
					} else if (cmd == "select_group") {
						group_insert(rs.include_group,name);
					} else if (cmd == "select_type") {
						type_insert(rs.include_type,name);
					}
				} else if (cmd == "game") {
					string game;
					string group;
					string type;
					string time;

					while (i < s.length() && !isspace(s[i])) {
						game += s[i];
						++i;
					}
	
					while (i < s.length() && isspace(s[i]))
						++i;

					if (i >= s.length() || s[i]!='"')
						goto err;
					++i;
	
					while (i < s.length() && s[i]!='"') {
						group += s[i];
						++i;
					}

					if (group.length()==0)
						group = GROUP_UNDEFINED;

					if (i >= s.length() || s[i]!='"')
						goto err;
					++i;

					while (i < s.length() && isspace(s[i]))
						++i;

					if (i >= s.length() || s[i]!='"')
						goto err;
					++i;
	
					while (i < s.length() && s[i]!='"') {
						type += s[i];
						++i;
					}

					if (type.length()==0)
						type = TYPE_UNDEFINED;

					if (i >= s.length() || s[i]!='"')
						goto err;
					++i;

					while (i < s.length() && isspace(s[i]))
						++i;

					while (i < s.length() && !isspace(s[i])) {
						time += s[i];
						++i;
					}

					while (i < s.length() && isspace(s[i]))
						++i;

					if (time.length() == 0 || i<s.length())
						goto err;

					if (!scan_game(gar,grp,typ,game,group,type,time))
						cerr << "warning: ignoring info for game " << game << endl;
				} else if (cmd == "select_neogeo") {
					rs.exclude_neogeo = string2tristate(value);
				} else if (cmd == "select_clone") {
					rs.exclude_clone = string2tristate(value);
				} else if (cmd == "select_bad") {
					rs.exclude_bad = string2tristate(value);
				} else if (cmd == "select_missing") {
					rs.exclude_missing = string2tristate(value);
				} else if (cmd == "select_vector") {
					rs.exclude_vector = string2tristate(value);
				} else if (cmd == "sort") {
					rs.sort_mode = string2sort(value);
					if (rs.sort_mode < 0)
						goto err_arg;
				} else if (cmd == "mode") {
					if (value == "list")
						rs.mode = mode_list;
					else if (value == "tile_3x2")
						rs.mode = mode_tile_3x2;
					else if (value == "tile_4x3")
						rs.mode = mode_tile_4x3;
					else if (value == "tile_5x4")
						rs.mode = mode_tile_5x4;
					else if (value == "text") /* compatibility with the old version */
						rs.mode = mode_list;
					else
						goto err_arg;
				} else if (cmd == "exit_press") {
					if (value == "1")
						rs.exit_count = 1;
					else if (value == "2")
						rs.exit_count = 2;
					else if (value == "3")
						rs.exit_count = 3;
					else
						goto err_arg;
				} else if (cmd == "event") {
					if (!text_key_in(value))
						goto err;
				} else if (cmd == "color") {
					if (!color_in(value))
						goto err;
				} else if (cmd == "idle_start") {
					sscanf(value.c_str(),"%d %d",&rs.idle_start,&rs.idle_start_rep);
				} else if (cmd == "last_game") {
					rs.last_game = value;
				} else if (cmd == "idle_move") {
					sscanf(value.c_str(),"%d %d",&rs.idle_move,&rs.idle_move_rep);
				} else if (cmd == "preview") {
					if (value == "snap")
						rs.preview = preview_snap;
					else if (value == "flyers")
						rs.preview = preview_flyers;
					else
						goto err_arg;
				} else if (cmd == "merge") {
					if (value == "no")
						rs.merge = merge_no;
					else if (value == "differential")
						rs.merge = merge_differential;
					else if (value == "parent")
						rs.merge = merge_parent;
					else if (value == "any")
						rs.merge = merge_any;
					else if (value == "disable")
						rs.merge = merge_disable;
					else
						goto err_arg;
				} else if (cmd == "crc_scan") {
					if (value == "no")
						rs.crc_scan = false;
					else if (value == "yes")
						rs.crc_scan = true;
					else
						goto err_arg;
				} else {
					cerr << "warning: ignoring command " << cmd << " in the configuration file " << file << endl;
				}

				continue;
			err:
				cerr << "error: in the configuration file " << file << endl;
				cerr << "error: at line '" << s <<  "'" << endl;
				exit(EXIT_FAILURE);

			err_arg:
				cerr << "error: in the configuration file " << file << endl;
				cerr << "error: at line '" << s <<  "'" << endl;
				cerr << "error: unknow argument " << value << endl;
				exit(EXIT_FAILURE);
			}
		}

		f.close();
	}

	for(game_set::const_iterator i=gar.begin();i!=gar.end();++i)
		if (i->group_get().length()==0)
			i->group_set(GROUP_UNDEFINED);

	group_insert(grp,GROUP_UNDEFINED);
	type_insert(typ,TYPE_UNDEFINED);

	if (rs.include_group.size() == 0)
		rs.include_group = grp;
	if (rs.include_type.size() == 0)
		rs.include_type = typ;
}

void save_grp(const menu_state& rs, const game_set& gar, const group_container& grp, const type_container& typ, const string& file) {
	// skip if no write access
	if (access(file.c_str(), F_OK) == 0 && access(file.c_str(), W_OK) != 0)
		return;

	ofstream f(file.c_str(), ios::out );
	if (!f) {
		cerr << "error: opening for writing the configuration file " << file << endl;
		exit(EXIT_FAILURE);
	}

	f << "# Events" << endl;
	f << "# event EVENT KEY0 KEY1" << endl;
	text_key_out_all(f);
	f << endl;

	f << "# Colors" << endl;
	f << "# color ITEM FOREGROUND BACKGROUND" << endl;
	f << "# Where FOREGROUND/BACKGROUND can be one of the following: " << endl;
	f << "#   black, blue, green, cyan, red, magenta, brown, lightgray, gray," << endl;
	f << "#   lightblue, lightgreen, lightcyan, lightred, lightmagenta, yellow, white" << endl;
	color_out_all(f);
	f << endl;

	f << "# Mode" << endl;
	f << "# mode (list | tile_3x2 | tile_4x3)" << endl;
	switch (rs.mode) {
		case mode_list: f << "mode list" << endl; break;
		case mode_tile_3x2: f << "mode tile_3x2" << endl; break;
		case mode_tile_4x3: f << "mode tile_4x3" << endl; break;
		case mode_tile_5x4: f << "mode tile_5x4" << endl; break;
	}
	f << endl;

	f << "# Exit press" << endl;
	f << "# exit_press (1 | 2 | 3)" << endl;
	f << "exit_press " << rs.exit_count << endl;
	f << endl;

	f << "# Merge" << endl;
	f << "# merge no | parent | differential | any | disable" << endl;
	f << "merge ";
	switch (rs.merge) {
		case merge_no : f << "no"; break;
		case merge_parent : f << "parent"; break;
		case merge_differential : f << "differential"; break;
		case merge_any : f << "any"; break;
		case merge_disable : f << "disable"; break;
	}
	f << endl;
	f << endl;

	f << "# Crc scan (only for mess)" << endl;
	f << "# crc_scan no | yes" << endl;
	f << "crc_scan ";
	if (rs.crc_scan)
		f << "yes";
	else
		f << "no";
	f << endl;
	f << endl;

	f << "# Idle" << endl;
	f << "# (idle_start | idle_move) SECOND_START SECOND_REPEAT" << endl;
	f << "idle_start " << rs.idle_start << " " << rs.idle_start_rep << endl;
	f << "idle_move " << rs.idle_move << " " << rs.idle_move_rep << endl;
	f << endl;

	f << "# Last game" << endl;
	f << "# last_game GAMENAME" << endl;
	if (rs.last_game.length()) {
		f << "last_game " << rs.last_game << endl;
		f << endl;
	}

	f << "# Sort mode" << endl;
	f << "# sort (group | parent | name | time | year | manufacturer | type | size)" << endl;
	f << "sort " << sort2string(rs.sort_mode) << endl;
	f << endl;

	f << "# Game selected" << endl;
	f << "# select_(neogeo | clone | bad | missing | vector) (include | exclude | exclude_not)" << endl;
	f << "select_neogeo " << tristate2string(rs.exclude_neogeo) << endl;
	f << "select_clone " << tristate2string(rs.exclude_clone) << endl;
	f << "select_bad " << tristate2string(rs.exclude_bad) << endl;
	f << "select_missing " << tristate2string(rs.exclude_missing) << endl;
	f << "select_vector " << tristate2string(rs.exclude_vector) << endl;
	f << endl;

	f << "# Preview" << endl;
	f << "# preview (snap | flyers)" << endl;
	f << "preview ";
	switch (rs.preview) {
		case preview_snap : f << "snap"; break;
		case preview_flyers : f << "flyers"; break;
	}
	f << endl;
	f << endl;

	f << "# Group set" << endl;
	f << "# group STRING" << endl;
	for(group_container::const_iterator i=grp.begin();i!=grp.end();++i) {
		if (*i!=GROUP_UNDEFINED) {
			f << "group \"" << *i << "\"" << endl;
		}
	}
	f << endl;

	f << "# Group selected" << endl;
	f << "# select_group STRING" << endl;
	for(group_container::const_iterator i=rs.include_group.begin();i!=rs.include_group.end();++i) {
		f << "select_group \"" << *i << "\"" << endl;
	}
	f << endl;

	f << "# Type set" << endl;
	f << "# type = STRING" << endl;
	for(group_container::const_iterator i=typ.begin();i!=typ.end();++i) {
		if (*i!=TYPE_UNDEFINED) {
			f << "type \"" << *i << "\"" << endl;
		}
	}
	f << endl;

	f << "# Type selected" << endl;
	f << "# select_type = STRING" << endl;
	for(group_container::const_iterator i=rs.include_type.begin();i!=rs.include_type.end();++i) {
		f << "select_type \"" << *i << "\"" << endl;
	}
	f << endl;

	f << "# Game list" << endl;
	f << "# game GROUP_STRING TYPE_STRING SECONDS_PLAYED" << endl;
	for(game_set::const_iterator i=gar.begin();i!=gar.end();++i) {
		if ((i->group_get().length()!=0 && i->group_get()!=GROUP_UNDEFINED) ||
			(i->type_get().length()!=0 && i->type_get()!=TYPE_UNDEFINED) ||
			i->time_get()!=0) {
			f << "game " << i->name_get();
			f << " \"";
			if (i->group_get()!=GROUP_UNDEFINED)
				f << i->group_get();
			f << "\"";
			f << " \"";
			if (i->type_get()!=TYPE_UNDEFINED)
				f << i->type_get();
			f << "\"";
			f << " " << i->time_get();
			f << endl;
		}
	}
	f << endl;

	f.close();
}

void scan_arg(string& snap, string& font, int argc, char* argv[]) {
	int c;
	while ((c=getopt(argc, argv, "s:f:")) != -1) {
		switch (c) {
			case 's' :
				snap = snap + ";" + optarg;
				break;
			case 'f' :
				font = optarg;
				break;
			case '?':
				printf("Unknown option %c\n", c);
				exit(1);
		}
	}
}

// --------------------------------------------------------------------------
// Run

int run_main(const string& font,const game_set& gar, const group_container& grp, const group_container& typ, unsigned orientation, double gamma, double brightness, menu_state& rs) {
	if (!text_sub_init(font.c_str(),orientation,gamma,brightness,rs.idle_start,rs.idle_start_rep,rs.idle_move,rs.idle_move_rep)) {
		cerr << "error: inizializing the video system" << endl;
		exit(EXIT_FAILURE);
	}

	bool done = false;
	int key = 0;

	while (!done) {
		key = run_menu(gar,rs, (orientation & ORIENTATION_FLIP_XY) != 0);
		switch (key) {
			case TEXT_KEY_HELP :
				run_help();
				break;
			case TEXT_KEY_GROUP :
				run_group(rs,grp);
				break;
			case TEXT_KEY_TYPE :
				run_type(rs,typ);
				break;
			case TEXT_KEY_EXCLUDE :
				run_exclude(rs);
				break;
			case TEXT_KEY_SORT :
				run_sort(rs);
				break;
			case TEXT_KEY_SETGROUP :
				if (rs.pgame)
					run_group_move(rs,grp);
				break;
			case TEXT_KEY_SETTYPE :
				if (rs.pgame)
					run_type_move(rs,typ);
				break;
			case TEXT_KEY_RUN_CLONE :
				if (rs.pgame) {
					run_clone(rs);
					if (rs.pclone)
						done = true;
				}
				break;
			case TEXT_KEY_IDLE_0 :
				if (rs.pgame) {
					rs.pclone = rs.pgame->clone_best_get();
					done = true;
				}
				break;
			case TEXT_KEY_ENTER :
				if (rs.pgame) {
					if (rs.exclude_clone == exclude)
						rs.pclone = rs.pgame->clone_best_get();
					else
						rs.pclone = rs.pgame;
					done = true;
				}
				break;
			case TEXT_KEY_ESC :
				done = true;
				break;
		}
	}

	text_sub_done();

	return key;
}

string snapshot_name_get(const string& snap_create, const string& name) {
	char path[256];
	sprintf(path,"%s%.8s.png", snap_create.c_str(),name.c_str());

	int snapno = 0;
	while (access(path,F_OK)==0) {
		sprintf(path,"%s%.4s%04d.png",snap_create.c_str(),name.c_str(),snapno++);
	}

	return path;
}

void run_game(emulator_t emulator, const game& g, bool ignore_error, const string& snap_create) {
	char cmd[256];
	time_t start,stop;
	string snapshot_rename_dir;
	string snapshot_create_file;
	string snapshot_rename_file;

	flush_cache();

	start = time(0);

	printf("%s, %s, %s\n", g.description_get().c_str(), g.manufacturer_get().c_str(), g.year_get().c_str());
	switch (emulator) {
		case emulator_mame: sprintf(cmd,"mame %s", g.name_get().c_str()); break;
		case emulator_mess: {
			if (g.software_get()) {
				const game* bios = g.bios_get();
				string stripped_name = string(g.name_get(),g.name_get().find('/') + 1);
				snapshot_create_file = snapshot_name_get(snap_create,bios->name_get());
				snapshot_rename_dir = snap_create + bios->name_get();
				snapshot_rename_file = snapshot_rename_dir + "/" + stripped_name + ".png";
				if (g.alias_get()) {
					sprintf(cmd,"mess %s -alias %s", bios->name_get().c_str(), stripped_name.c_str());
				} else {
					sprintf(cmd,"mess %s -cart %s", bios->name_get().c_str(), stripped_name.c_str());
				}
			} else {
				sprintf(cmd,"mess %s", g.name_get().c_str());
			}
		}
		break;
	}
	printf("%s\n",cmd);

	__djgpp_exception_toggle();
	int r = system(cmd);
	__djgpp_exception_toggle();

	stop = time(0);
	if (stop - start > 8) {
		g.time_set( g.time_get() + stop - start );
	}

	if (!ignore_error) {
		if (r == -1) {
			cerr << "MAME terminate with DOS error " << errno << endl;
			sleep(5);
		} else if (r & 0xFF != 0) {
			cerr << "MAME terminate with error " << (r & 0xFF) << endl;
			sleep(5);
		} else if ((r >> 8) & 0xFF != 0) {
			cerr << "MAME terminate with signal " << ((r >> 8) & 0xFF) << endl;
			sleep(5);
		}
	}

	switch (emulator) {
		case emulator_mame:
			break;
		case emulator_mess: {
			if (g.software_get()) {
				if (access(snapshot_create_file.c_str(),F_OK)==0) {
					mkdir(snapshot_rename_dir.c_str(),S_IWUSR);
					rename(snapshot_create_file.c_str(), snapshot_rename_file.c_str());
				}
			}
		}
		break;
	}

	text_idle_time_reset();

	// if return with success
	if ((r & 0xFF)==0)
		text_idle_repeat_reset();

	flush_cache();
}

int run_all(menu_state& rs, const string& font, const game_set& gar, const group_container& grp, const type_container& typ, unsigned orientation, double gamma, double brightness) {
	bool done = false;
	int key = 0;

	rs.pos_base = 0;
	rs.pos_rel = 0;
	rs.pgame = 0;
	rs.pclone = 0;
	rs.fast = "";

	while (!done) {
		key = run_main(font,gar,grp,typ,orientation,gamma,brightness,rs);
		switch (key) {
			case TEXT_KEY_ESC :
				done = true;
				break;
			case TEXT_KEY_ENTER :
			case TEXT_KEY_IDLE_0 :
			case TEXT_KEY_RUN_CLONE :
				if (!rs.pclone)
					rs.pclone = rs.pgame;
				if (rs.pclone) {
					// save before
					save_grp(rs,gar,grp,typ,FILE_CFG);
					run_game(rs.emulator,*rs.pclone, key == TEXT_KEY_IDLE_0,rs.snap_create_path);
					// save after
					save_grp(rs,gar,grp,typ,FILE_CFG);
				}
				break;
		}
	}

	return key;
}

int main(int argc, char* argv[]) {
	game_set gar;
	group_container grp;
	type_container typ;
	string font = FILE_FONT;
	unsigned orientation = 0;
	double gamma = 1;
	double brightness = 1;
	menu_state rs;

	if (access(FILE_MAME,F_OK)!=0 && access(FILE_MESS,F_OK)==0)
		rs.emulator = emulator_mess;
	else
		rs.emulator = emulator_mame;

	rs.idle_start = 0;
	rs.idle_start_rep = 0;
	rs.idle_move = 40;
	rs.idle_move_rep = 5;

	rs.sort_mode = sort_by_root_name;
	rs.exclude_neogeo = include;
	rs.exclude_clone = exclude;
	rs.exclude_bad = exclude;
	rs.exclude_missing = include;
	rs.exclude_vector = include;

	rs.mode = mode_list;
	rs.exit_count = 1;
	rs.merge = merge_differential;

	rs.preview = preview_snap;
	rs.crc_scan = true;

	srand(time(0));

	bool text = false;
	switch (rs.emulator) {
		case emulator_mame : text = text_init(FILE_CFG_MAME); break;
		case emulator_mess : text = text_init(FILE_CFG_MESS); break;
	}
	if (!text)
		exit(1);

	switch (rs.emulator) {
		case emulator_mame : scan_info(gar,FILE_INFO,FILE_MAME); break;
		case emulator_mess : scan_info(gar,FILE_INFO,FILE_MESS); break;
	}

	scan_cfg(rs,gar,orientation,gamma,brightness);

	if (rs.emulator == emulator_mess) {
		game_container gac;
		crc_list_set cls;

		if (rs.crc_scan)
			crc_scan(cls,gar,rs.crc_path);

		// if cls is empty the crc scan is automatically disabled
		scan_alias(gar,gac,FILE_CFG_MESS,cls);
		scan_software(gac,gar,cls);

		// any duplicate alias/software is rejected in the insertion
		// in the game_set. the first is used.
		for(game_container::const_iterator i=gac.begin();i!=gac.end();++i)
			gar.insert(*i);
	}

	gar.sync();

	scan_grp(rs,gar,grp,typ,FILE_CFG);

	scan_arg(rs.snap_path,font,argc,argv);

	sound_signal();

	run_all(rs,font,gar,grp,typ,orientation,gamma,brightness);

	save_grp(rs,gar,grp,typ,FILE_CFG);

	text_done();

	return EXIT_SUCCESS;
}

