/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "bitmap.h"

#include <mem.h>
#include <assert.h>

struct bitmap* bitmap_alloc(unsigned x, unsigned y, unsigned bit) {
	struct bitmap* bmp = (struct bitmap*)malloc(sizeof(struct bitmap));
	assert( bmp );

	bmp->size_x = x;
	bmp->size_y = y;
	bmp->bytes_per_pixel = (bit + 7) / 8;
	bmp->bytes_per_scanline = (bmp->size_x * bmp->bytes_per_pixel + 3) & ~0x3;
	bmp->bytes = bmp->bytes_per_scanline * bmp->size_y;
	bmp->data = malloc( bmp->bytes );
	assert( bmp->data );

	return bmp;
}

void bitmap_free(struct bitmap* bmp) {
	if (bmp)
		free(bmp->data);
	free(bmp);
}

uint8* bitmap_line(struct bitmap* bmp, unsigned line) {
	assert( bmp );
	switch (bmp->bytes_per_scanline) {
		case 8 :
			return bmp->data + line * 8;
		case 16 :
			return bmp->data + line * 16;
		default:
			return bmp->data + line * bmp->bytes_per_scanline;
	}
}

void bitmap_putpixel(struct bitmap* bmp, unsigned x, unsigned y, unsigned v) {
	uint8* ptr = bmp->data + x * bmp->bytes_per_pixel + y * bmp->bytes_per_scanline;
	switch (bmp->bytes_per_pixel) {
		case 1 :
			ptr[0] = v;
		break;
		case 2 :
			ptr[0] = v;
			ptr[1] = v >> 8;
		break;
		case 3 :
			ptr[0] = v;
			ptr[1] = v >> 8;
			ptr[2] = v >> 16;
		break;
		case 4 :
			ptr[0] = v;
			ptr[1] = v >> 8;
			ptr[2] = v >> 16;
			ptr[3] = v >> 24;
		break;
	};
}

#define FASTEXIT_Y_CHECK 16

int bitmap_orientation(struct bitmap* bmp, unsigned orientation_mask, int (*fast_exit_handler)(void)) {
	if (orientation_mask & ORIENTATION_FLIP_XY) {
		struct bitmap* newbmp;

		/* new data */
		newbmp = bitmap_alloc(bmp->size_y,bmp->size_x,bmp->bytes_per_pixel * 8);
		assert( newbmp );

		{
			unsigned size_x = newbmp->size_x;
			unsigned size_y = newbmp->size_y;
			unsigned src_bytes_per_scanline = bmp->bytes_per_scanline;
			unsigned dst_bytes_per_scanline = newbmp->bytes_per_scanline;
			uint8* src = (uint8*)bmp->data;
			uint8* dst = (uint8*)newbmp->data;

			if (bmp->bytes_per_pixel == 1) {
				unsigned y;
				for(y=0;y<size_y;++y) {
					uint8* srcline = src;
					uint8* dstline = dst;
					unsigned x;
					for(x=0;x<size_x;++x) {
						*dstline = *srcline;
						dstline += 1;
						srcline += src_bytes_per_scanline;
					}
					dst += dst_bytes_per_scanline;
					src += 1;
					if (y % FASTEXIT_Y_CHECK == 0 && fast_exit_handler && fast_exit_handler())
						return -1;
				}
			} else if (bmp->bytes_per_pixel == 2) {
				unsigned y;
				for(y=0;y<size_y;++y) {
					uint8* srcline = src;
					uint8* dstline = dst;
					unsigned x;
					for(x=0;x<size_x;++x) {
						dstline[0] = srcline[0];
						dstline[1] = srcline[1];
						dstline += 2;
						srcline += src_bytes_per_scanline;
					}
					dst += dst_bytes_per_scanline;
					src += 2;
					if (y % FASTEXIT_Y_CHECK == 0 && fast_exit_handler && fast_exit_handler())
						return -1;
				}
			} else if (bmp->bytes_per_pixel == 3) {
				unsigned y;
				for(y=0;y<size_y;++y) {
					uint8* srcline = src;
					uint8* dstline = dst;
					unsigned x;
					for(x=0;x<size_x;++x) {
						dstline[0] = srcline[0];
						dstline[1] = srcline[1];
						dstline[2] = srcline[2];
						dstline += 3;
						srcline += src_bytes_per_scanline;
					}
					dst += dst_bytes_per_scanline;
					src += 3;
					if (y % FASTEXIT_Y_CHECK == 0 && fast_exit_handler && fast_exit_handler())
						return -1;
				}
			} else if (bmp->bytes_per_pixel == 4) {
				unsigned y;
				for(y=0;y<size_y;++y) {
					uint8* srcline = src;
					uint8* dstline = dst;
					unsigned x;
					for(x=0;x<size_x;++x) {
						/* dstline and srcline are always aligned at 4 bytes */
						*(uint32*)dstline = *(uint32*)srcline;
						dstline += 4;
						srcline += src_bytes_per_scanline;
					}
					dst += dst_bytes_per_scanline;
					src += 4;
					if (y % FASTEXIT_Y_CHECK == 0 && fast_exit_handler && fast_exit_handler())
						return -1;
				}
			}
		}

		/* copy */
		free(bmp->data);
		*bmp = *newbmp;
	}

	if (orientation_mask & ORIENTATION_MIRROR_Y) {
		unsigned bytes_per_scanline = bmp->bytes_per_scanline;
		uint8* y0 = bitmap_line(bmp,0);
		uint8* y1 = bitmap_line(bmp,bmp->size_y - 1);
		void* buf = malloc(bytes_per_scanline);
		for(;y0<y1; y0 += bytes_per_scanline, y1 -= bytes_per_scanline) {
			memcpy(buf,y0,bytes_per_scanline);
			memcpy(y0,y1,bytes_per_scanline);
			memcpy(y1,buf,bytes_per_scanline);
		}
		free(buf);
	}

	if (orientation_mask & ORIENTATION_MIRROR_X) {
		if (bmp->bytes_per_pixel == 1) {
			unsigned y;
			for(y=0;y<bmp->size_y;++y) {
				uint8* x0 = (uint8*)bitmap_line(bmp,y);
				uint8* x1 = x0 + bmp->size_x - 1;
				for(;x0<x1;++x0,--x1) {
					uint8 t = *x0;
					*x0 = *x1;
					*x1 = t;
				}
			}
		} else {
			unsigned y;
			for(y=0;y<bmp->size_y;++y) {
				uint16* x0 = (uint16*)bitmap_line(bmp,y);
				uint16* x1 = x0 + bmp->size_x - 1;
				for(;x0<x1;++x0,--x1) {
					uint16 t = *x0;
					*x0 = *x1;
					*x1 = t;
				}
			}
		}
	}

	return 0;
}

struct color_node {
	unsigned index;
	unsigned count;
};

#define COUNT_SORT_BIT_MAX 8
#define COUNT_SORT_MAX (1U << COUNT_SORT_BIT_MAX)

static unsigned count_sort[COUNT_SORT_MAX];

/*
static void countsort(struct color_node* indexin[], struct color_node* indexout[], unsigned bit, unsigned skipbit) {
	unsigned max = 1 << bit;
	unsigned mask = max - 1;
	unsigned i;

	for(i=0;i<=max;i++)
		count_sort[i] = 0;

	for(i=0;i<BITMAP_INDEX_MAX;i++) {
		unsigned j = (indexin[i]->count >> skipbit) & mask;
		count_sort[j+1]++;
	}

	for(i=1;i<max;i++)
		count_sort[i] += count_sort[i-1];

	for(i=0;i<BITMAP_INDEX_MAX;i++) {
		unsigned j = (indexin[i]->count >> skipbit) & mask;
		indexout[count_sort[j]] = indexin[i];
		count_sort[j]++;
	}
}
*/

static void countsort80(struct color_node* indexin[], struct color_node* indexout[]) {
	unsigned i;

	for(i=0;i<=256;i++)
		count_sort[i] = 0;

	for(i=0;i<BITMAP_INDEX_MAX;i++) {
		unsigned j = indexin[i]->count & 0xFF;
		count_sort[j+1]++;
	}

	for(i=1;i<256;i++)
		count_sort[i] += count_sort[i-1];

	for(i=0;i<BITMAP_INDEX_MAX;i++) {
		unsigned j = indexin[i]->count & 0xFF;
		indexout[count_sort[j]] = indexin[i];
		count_sort[j]++;
	}
}

static void countsort88(struct color_node* indexin[], struct color_node* indexout[]) {
	unsigned i;

	for(i=0;i<=256;i++)
		count_sort[i] = 0;

	for(i=0;i<BITMAP_INDEX_MAX;i++) {
		unsigned j = (indexin[i]->count >> 8) & 0xFF;
		count_sort[j+1]++;
	}

	for(i=1;i<256;i++)
		count_sort[i] += count_sort[i-1];

	for(i=0;i<BITMAP_INDEX_MAX;i++) {
		unsigned j = (indexin[i]->count >> 8) & 0xFF;
		indexout[count_sort[j]] = indexin[i];
		count_sort[j]++;
	}
}

unsigned bitmap_reduction(unsigned* convert, video_color* palette, unsigned size, const struct bitmap* bmp, int (*fast_exit_handler)(void)) {
	unsigned i,y;
	unsigned res_size;
	static struct color_node map[BITMAP_INDEX_MAX];
	static struct color_node* index1[BITMAP_INDEX_MAX];
	static struct color_node* index2[BITMAP_INDEX_MAX];

	(void)fast_exit_handler;

	assert( bmp->bytes_per_pixel == 3 );

	/* clear all */
	for(i=0;i<BITMAP_INDEX_MAX;++i) {
		map[i].count = 0;
		map[i].index = 0;
		index1[i] = map + i;
	}

	/* count */
	for(y=0;y<bmp->size_y;++y) {
		unsigned x;
		uint8* line = (uint8*)bitmap_line((struct bitmap* )bmp,y);
		for(x=0;x<bmp->size_x;++x) {
			unsigned index;
			index = BITMAP_COLOR_TO_INDEX( line[0], line[1], line[2] );
			++map[index].count;
			line += 3;
		}
	}

	/* sort */
	countsort80(index1,index2);
	countsort88(index2,index1);

	/* create palette */
	for(i=0;i<size && index1[BITMAP_INDEX_MAX - i - 1]->count;++i) {
		unsigned subindex = BITMAP_INDEX_MAX - i - 1;
		unsigned index = index1[subindex] - map;
		index1[subindex]->index = i + 1;
		palette[i].red = BITMAP_INDEX_TO_RED( index );
		palette[i].green = BITMAP_INDEX_TO_GREEN( index );
		palette[i].blue = BITMAP_INDEX_TO_BLUE( index );
		palette[i].alpha = 0;
	}
	res_size = i;

	/* make index table */
	for(i=0;i<BITMAP_INDEX_MAX;++i) {
		if (map[i].index) {
			convert[i] = map[i].index - 1;
		} else if (map[i].count) {
			int red = BITMAP_INDEX_TO_RED( i );
			int green = BITMAP_INDEX_TO_GREEN( i );
			int blue = BITMAP_INDEX_TO_BLUE( i );
			long diff = (red-palette[0].red)*(red-palette[0].red) + (green-palette[0].green)*(green-palette[0].green) + (blue-palette[0].blue)*(blue-palette[0].blue);
			unsigned best = 0;
			unsigned j;

			for(j=1;j<res_size;++j) {
				long new_diff = (red-palette[j].red)*(red-palette[j].red) + (green-palette[j].green)*(green-palette[j].green) + (blue-palette[j].blue)*(blue-palette[j].blue);
				if (new_diff < diff) {
					best = j;
					diff = new_diff;
				}
			}
			convert[i] = best;
		} else {
			convert[i] = 0;
		}
	}

	return res_size;
}
