/*
 * This file is part of the AdvanceMAME project
 *
 * Copyright (C) 1999-2000 Andrea Mazzoleni
 *
 * The contents of this file are free for non-commercial use.
 * Any commercial use is forbidden.
 */

#include "scrtweak.h"

#include <dos.h>
#include <assert.h>
#include <io.h>

/* Does the initialization to make the VGA ready to
 * accept any combination of configuration register settings.
 * This involves enabling writes to index 0 to 7 of the CRT controller
 * (port 0x3d4), by clearing the most significant bit (bit 7) of index
 * 0x11.
 */
void tweak_ready_vga_regs(void) {
	unsigned value;
	/* Ensure CRT regs. 0-7 are writable! */
	/* That is, clear bit 7 of port 0x3D4, index 0x11. */
	outportb(0x3d4,0x11);
	value = inportb(0x3d5) & 0x7F;
	outportw(0x3d4,0x11 | (value << 8));
}

void tweak_display_enable(void) {
	/* Reenable display data */
	outportb(0x3c0, 0x20);
}

/* Sets a single register according to the contents of the
 * passed tweak_register structure.
 */
void tweak_reg_out(unsigned value, unsigned port, int index) {
	switch (port)   {
		case TWEAK_ATTRCON_ADDR:
			/* reset read/write flip-flop */
			inportb(TWEAK_STATUS_ADDR);
			/* ensure VGA output is enabled */
			outportb(TWEAK_ATTRCON_ADDR, index | 0x20);
			outportb(TWEAK_ATTRCON_ADDR, value);
			break;
		case TWEAK_SEQ_ADDR:
			if (index == 1) {
				/* Reset the sequencer if the clock polarity settings */
				/* are being accessed. */
				outportw(TWEAK_SEQ_ADDR, 0x0100);
				outportw(TWEAK_SEQ_ADDR, (value << 8) | 1);
				outportw(TWEAK_SEQ_ADDR, 0x0300);
				break;
			}
			/* NOT break */
		case TWEAK_GRACON_ADDR:
		case TWEAK_CRTC_ADDR:
		case TWEAK_CHIPSTECH_ADDR:
			/* index to port, value to port+1 */
			outportw(port, index | value << 8);
			break;
		case TWEAK_MISC_ADDR:
		case TWEAK_VGAENABLE_ADDR:
		default:
			/* directly to the port */
			outportb(port, value);
			break;
	}
}

/* Inputs Register.value to the associated hardware register.
 * It takes care of handling recognized special cases,
 * like some VGA ports, correctly.
 */
unsigned tweak_reg_in(unsigned port, int index) {
	switch (port) {
		case TWEAK_MISC_ADDR:
			/* 0x3c2 is write-only, reading */
			/* must be mapped to 0x3cc */
			return inportb(0x3cc);
			break;
		case TWEAK_ATTRCON_ADDR:
			/* This 1 is odd.  First do a read to */
			/* reset the index/data flip-flop. */
			/* Then give it the index, but: */
			/* set bit 5!  If cleared, VGA */
			/* output is disabled! */
			inportb(TWEAK_STATUS_ADDR);
			outportb(TWEAK_ATTRCON_ADDR, index | (1 << 5));
			return inportb(TWEAK_ATTRCON_ADDR+1);
			break;
		case TWEAK_SEQ_ADDR:
		case TWEAK_GRACON_ADDR:
		case TWEAK_CRTC_ADDR:
		case TWEAK_CHIPSTECH_ADDR:
			/* These 3 are similar.  Give it the */
			/* register index, then read the */
			/* byte from port+1. */
			outportb(port, index);
			return inportb(port+1);
			break;
		case TWEAK_VGAENABLE_ADDR:
		default:
			/* Default is to read the byte directly from the port */
			return inportb(port);
	}
}

int tweak_reg_write(unsigned value, unsigned port, int index, void* arg) {
	(void)arg;
	tweak_reg_out(value,port,index);
	return 0;
}

unsigned tweak_reg_read(unsigned port, int index, void* arg) {
	(void)arg;
	return tweak_reg_in(port,index);
}


/* Sets n registers according to the array pointed to by r.
 * First, indexes 0-7 of the CRT controller are enabled for writing.
 */
void tweak_reg_array_out(const tweak_register *r, int n) {
	tweak_ready_vga_regs();

	while (n--) {
		tweak_reg_out(r->value,r->port,r->index);
		++r;
	}
}

/* Bit mask a register */
static __inline__ unsigned mask(unsigned value, unsigned rbit, unsigned len) {
	return (value >> rbit) & ((1 << len)-1);
}

/* Compute an end value */
static __inline__ unsigned end_value(unsigned start, unsigned end_nibble, unsigned end_bits) {
	unsigned end;
	unsigned end_mask = (1 << end_bits) - 1;
	assert( (end_nibble & ~end_mask) == 0 );
	if ((start & end_mask) > end_nibble) {
		end = start + end_mask + 1;
	} else {
		end = start;
	}
	return (end & ~end_mask) | end_nibble;
}

/* Compile tweak_mode_info strcture with video mode information */
void tweak_crtc_get(struct tweak_crtc* crtc, unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {

	unsigned GC05 = getreg(TWEAK_GRACON_ADDR, 0x05, arg);
	unsigned GC06 = getreg(TWEAK_GRACON_ADDR, 0x06, arg);
	unsigned AC10 = getreg(TWEAK_ATTRCON_ADDR, 0x10, arg);
	unsigned SEQ01 = getreg(TWEAK_SEQ_ADDR, 0x1, arg);
	unsigned SEQ04 = getreg(TWEAK_SEQ_ADDR, 0x4, arg);
	unsigned MISC = getreg(TWEAK_MISC_ADDR, -1, arg);
	unsigned CRTC00 = getreg(TWEAK_CRTC_ADDR, 0x0, arg);
	unsigned CRTC01 = getreg(TWEAK_CRTC_ADDR, 0x1, arg);
	unsigned CRTC02 = getreg(TWEAK_CRTC_ADDR, 0x2, arg);
	unsigned CRTC03 = getreg(TWEAK_CRTC_ADDR, 0x3, arg);
	unsigned CRTC04 = getreg(TWEAK_CRTC_ADDR, 0x4, arg);
	unsigned CRTC05 = getreg(TWEAK_CRTC_ADDR, 0x5, arg);
	unsigned CRTC06 = getreg(TWEAK_CRTC_ADDR, 0x6, arg);
	unsigned CRTC07 = getreg(TWEAK_CRTC_ADDR, 0x07, arg);
	unsigned CRTC09 = getreg(TWEAK_CRTC_ADDR, 0x09, arg);
	unsigned CRTC10 = getreg(TWEAK_CRTC_ADDR, 0x10, arg);
	unsigned CRTC12 = getreg(TWEAK_CRTC_ADDR, 0x12, arg);
	unsigned CRTC11 = getreg(TWEAK_CRTC_ADDR, 0x11, arg);
	unsigned CRTC13 = getreg(TWEAK_CRTC_ADDR, 0x13, arg);
	unsigned CRTC14 = getreg(TWEAK_CRTC_ADDR, 0x14, arg);
	unsigned CRTC15 = getreg(TWEAK_CRTC_ADDR, 0x15, arg);
	unsigned CRTC16 = getreg(TWEAK_CRTC_ADDR, 0x16, arg);
	unsigned CRTC17 = getreg(TWEAK_CRTC_ADDR, 0x17, arg);

	int temp;

	crtc->dotclock_master = mask(MISC,2,2);
	switch (crtc->dotclock_master) {
		case 0 :
			crtc->dot_clock = TWEAK_DOTCLOCK_LOW;
			break;
		case 1 :
			crtc->dot_clock = TWEAK_DOTCLOCK_HIGH;
			break;
		default:
			crtc->dot_clock = 0;
	}
	switch (mask(SEQ01,3,1)) {
		case 0:
			crtc->dotclock_middle = 0;
			break;
		case 1:
			crtc->dot_clock /= 2;
			crtc->dotclock_middle = 1;
			break;
	}
	switch (mask(SEQ01,0,1)) {
		case 0 :
			crtc->char_size_x = 9;
			break;
		case 1 :
			crtc->char_size_x = 8;
			break;
		default:
			crtc->char_size_x = 0;
	}
	switch (mask(GC05,5,2)) {
		case 0:
			crtc->bits_per_pixel = 4; /* might also be COLOR2 !!! */
			crtc->pixels_per_clock = 8;
			crtc->bytes_per_clock = 1;
			break;
		case 1:
			crtc->bits_per_pixel = 2;
			crtc->pixels_per_clock = 16;
			crtc->bytes_per_clock = 2;
			break;
		case 2:
			crtc->bits_per_pixel = 8;
			crtc->pixels_per_clock = 4;
			crtc->bytes_per_clock = 4;
			break;
		default:
			crtc->bits_per_pixel = 0;
			crtc->pixels_per_clock = 0;
			crtc->bytes_per_clock = 0;
	}

	temp = ((GC06 & 1) != 0);
	if (temp == ((AC10 & 1) != 0))
		crtc->is_graphics_mode = temp ? 1 : 0;
	else {
		crtc->is_graphics_mode = -1;
	}

	if (!crtc->is_graphics_mode) {
		crtc->bits_per_pixel = 0;
		crtc->pixels_per_clock = crtc->char_size_x;
		crtc->bytes_per_clock = 2;
		crtc->dots_per_clock = crtc->char_size_x;
	} else {
		crtc->dots_per_clock = 8;
	}

	crtc->is_linear = mask(SEQ04,3,1);
#if 1
	if (mask(CRTC14,6,1))
		crtc->memory_mode = 4; /* dword mode */
	else if (mask(CRTC17,6,1)==0)
		crtc->memory_mode = 2; /* word mode */
	else
		crtc->memory_mode = 1; /* byte mode */
#else
	if (mask(CRTC17,3,1))
		crtc->memory_mode = 4;
	else
		crtc->memory_mode = 2;
#endif
	crtc->clocks_per_scanline = CRTC13 * 2 * crtc->memory_mode / crtc->bytes_per_clock;

	crtc->hde = CRTC01 + 1;
	crtc->vde = (CRTC12 | (mask(CRTC07,1,1) << 8) | (mask(CRTC07,6,1) << 9)) + 1;

	crtc->size_x = crtc->hde * crtc->pixels_per_clock;
	crtc->virtual_x = crtc->clocks_per_scanline * crtc->pixels_per_clock;
	if (crtc->is_graphics_mode && !crtc->is_linear)
		crtc->virtual_x *= 4;

	crtc->bytes_per_scanline = CRTC13 * 2 * crtc->memory_mode;

	crtc->char_size_y = mask(CRTC09,0,4) + 1;
	crtc->scanlines_per_line = crtc->char_size_y;
	crtc->is_doublescan = mask(CRTC09,7,1);
	if (crtc->is_doublescan)
		crtc->scanlines_per_line *= 2;

	if (crtc->scanlines_per_line != 0) {
		crtc->size_y = crtc->vde / crtc->scanlines_per_line;
		crtc->partial_y = crtc->vde % crtc->scanlines_per_line;
	} else {
		crtc->size_y = 0;
		crtc->partial_y = 0;
	}
	if (crtc->bytes_per_scanline != 0) {
		crtc->virtual_y = (crtc->is_graphics_mode ? 65536L : 32768L) / crtc->bytes_per_scanline;
	} else {
		crtc->virtual_y = 0;
	}

	crtc->ht = CRTC00 + 5;
	crtc->vt = (CRTC06 | (mask(CRTC07,0,1) << 8) | (mask(CRTC07,5,1) << 9)) + 2;

	crtc->vrs = CRTC10 | (mask(CRTC07,2,1) << 8) | (mask(CRTC07,7,1) << 9);
	crtc->vre = end_value(crtc->vrs,mask(CRTC11,0,4),4);

	crtc->vbs = CRTC15 | (mask(CRTC07,3,1) << 8) | (mask(CRTC09,5,1) << 9);
	crtc->vbe = end_value(crtc->vbs,mask(CRTC16,0,8),8);
	crtc->vbe7 = end_value(crtc->vbs,mask(CRTC16,0,7),7);

	crtc->hbs = CRTC02;
	crtc->hbe = end_value(crtc->hbs,mask(CRTC03,0,5) | mask(CRTC05,7,1) << 5, 6);

	crtc->hrs = CRTC04;
	crtc->hre = end_value(crtc->hrs,mask(CRTC05,0,5),5);

	crtc->hsync = mask(MISC,6,1);
	crtc->vsync = mask(MISC,7,1);

	crtc->horz_clock = (double)crtc->dot_clock / (crtc->ht*crtc->dots_per_clock);
	crtc->vert_clock = (double)crtc->dot_clock / (crtc->ht*crtc->dots_per_clock*crtc->vt);
	if (crtc->dots_per_clock)
		crtc->pixel_clock = crtc->dot_clock * crtc->pixels_per_clock / crtc->dots_per_clock;
	else
		crtc->pixel_clock = 0;

	if (!crtc->is_graphics_mode) {
		crtc->size_y *= crtc->scanlines_per_line;
		crtc->virtual_y *= crtc->scanlines_per_line;
	}
}

/* Set a register with mask */
static int setregmask(unsigned value, unsigned rbit, unsigned lbit, unsigned port, int index, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	assert( lbit <= 8 );
	if (lbit != 8) {
		unsigned m = getreg(port,index,arg);
		m &= ~(((1 << lbit) - 1) << rbit);
		m |= value << rbit;
		return setreg(m,port,index,arg);
	} else {
		return setreg(value,port,index,arg);
	}
}

/* Get a register with mask */
static unsigned getregmask(unsigned rbit, unsigned lbit, unsigned port, int index, unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	unsigned m = getreg(port,index,arg);
	m >>= rbit;
	m &= (1 << lbit) - 1;
	return m;
}

int tweak_htt_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	(void)getreg;
	value -= 5;
	return setreg(value, TWEAK_CRTC_ADDR, 0x0, arg);
}

int tweak_hde_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	(void)getreg;
	value -= 1;
	return setreg(value, TWEAK_CRTC_ADDR, 0x1, arg);
}

unsigned tweak_hde_get(unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	unsigned value = getreg(TWEAK_CRTC_ADDR, 0x1, arg);
	value += 1;
	return value;
}

int tweak_hbs_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	(void)getreg;
	return setreg(value, TWEAK_CRTC_ADDR, 0x2, arg);
}

unsigned tweak_hbs_get(unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	unsigned value = 0;
	value |= getreg(TWEAK_CRTC_ADDR, 0x2, arg);
	return value;
}

int tweak_hbe_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	setregmask( mask(value,0,5), 0, 5, TWEAK_CRTC_ADDR, 0x3, setreg, getreg, arg);
	setregmask( mask(value,5,1), 7, 1, TWEAK_CRTC_ADDR, 0x5, setreg, getreg, arg);
	return 0;
}

unsigned tweak_hbe_get(unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	unsigned value = 0;
	value |= getregmask( 0, 5, TWEAK_CRTC_ADDR, 0x3, getreg, arg);
	value |= getregmask( 7, 1, TWEAK_CRTC_ADDR, 0x5, getreg, arg) << 5;
	return value;
}

int tweak_hrs_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	(void)getreg;
	return setreg(value, TWEAK_CRTC_ADDR, 0x4, arg);
}

int tweak_hre_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	return setregmask( mask(value,0,5), 0, 5, TWEAK_CRTC_ADDR, 0x5, setreg, getreg, arg);
}

int tweak_vtt_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	value -= 2;
	setregmask( mask(value,0,8), 0, 8, TWEAK_CRTC_ADDR, 0x6, setreg, getreg, arg);
	setregmask( mask(value,8,1), 0, 1, TWEAK_CRTC_ADDR, 0x7, setreg, getreg, arg);
	setregmask( mask(value,9,1), 5, 1, TWEAK_CRTC_ADDR, 0x7, setreg, getreg, arg);
	return 0;
}

int tweak_vde_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	value -= 1;
	setregmask( mask(value,0,8), 0, 8, TWEAK_CRTC_ADDR, 0x12, setreg, getreg, arg);
	setregmask( mask(value,8,1), 1, 1, TWEAK_CRTC_ADDR, 0x7, setreg, getreg, arg);
	setregmask( mask(value,9,1), 6, 1, TWEAK_CRTC_ADDR, 0x7, setreg, getreg, arg);
	return 0;
}

unsigned tweak_vde_get(unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	unsigned value = 0;
	value |= getregmask( 0, 8, TWEAK_CRTC_ADDR, 0x12, getreg, arg);
	value |= getregmask( 1, 1, TWEAK_CRTC_ADDR, 0x7, getreg, arg) << 8;
	value |= getregmask( 6, 1, TWEAK_CRTC_ADDR, 0x7, getreg, arg) << 9;
	value += 1;
	return value;
}

int tweak_vbs_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	setregmask( mask(value,0,8), 0, 8, TWEAK_CRTC_ADDR, 0x15, setreg, getreg, arg);
	setregmask( mask(value,8,1), 3, 1, TWEAK_CRTC_ADDR, 0x7, setreg, getreg, arg);
	setregmask( mask(value,9,1), 5, 1, TWEAK_CRTC_ADDR, 0x9, setreg, getreg, arg);
	return 0;
}

/*
 * note:
 *   Set a value compatible for 7 and 8 bit version of blank end.
 *   VGA doc report use of 7 bit, but other documentations report that many
 *   SVGA use all 8 bit.
 *   Call after vertical_blank_start_set.
 */
int tweak_vbe_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	/* get first 8 bit of start */
	unsigned start8 = getreg(TWEAK_CRTC_ADDR, 0x15, arg);
	/* compute first 8 bit of end, compatible also for 7 bit regs */
	unsigned end8 = end_value(start8, value & 0x7F, 7);
	return setregmask( mask(end8,0,8), 0, 8, TWEAK_CRTC_ADDR, 0x16, setreg, getreg, arg);
}

int tweak_vrs_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	setregmask( mask(value,0,8), 0, 8, TWEAK_CRTC_ADDR, 0x10, setreg, getreg, arg);
	setregmask( mask(value,8,1), 2, 1, TWEAK_CRTC_ADDR, 0x7, setreg, getreg, arg);
	setregmask( mask(value,9,1), 7, 1, TWEAK_CRTC_ADDR, 0x7, setreg, getreg, arg);
	return 0;
}

int tweak_vre_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	return setregmask( mask(value,0,4), 0, 4, TWEAK_CRTC_ADDR, 0x11, setreg, getreg, arg);
}

int tweak_doublescan_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	return setregmask( mask(value,0,1), 7, 1, TWEAK_CRTC_ADDR, 0x9, setreg, getreg, arg);
}

int tweak_dotclock_middle_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	return setregmask( mask(value,0,1), 3, 1, TWEAK_SEQ_ADDR, 0x1, setreg, getreg, arg);
}

int tweak_masterclock_input_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	return setregmask( mask(value,0,2), 2, 2, TWEAK_MISC_ADDR, -1, setreg, getreg, arg);
}

int tweak_char_size_x_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	if (value==8)
		value = 1;
	else if (value==9)
		value = 0;
	else
		return -1;
	return setregmask( mask(value,0,1), 0, 1, TWEAK_SEQ_ADDR, 0x1, setreg, getreg, arg);
}

int tweak_char_size_y_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	if (value<1 || value>32)
		return -1;
	value -= 1;
	return setregmask( mask(value,0,4), 0, 4, TWEAK_CRTC_ADDR, 0x9, setreg, getreg, arg);
}

int tweak_hsync_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	return setregmask( mask(value,0,1), 6, 1, TWEAK_MISC_ADDR, -1, setreg, getreg, arg);
}

int tweak_vsync_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	return setregmask( mask(value,0,1), 7, 1, TWEAK_MISC_ADDR, -1, setreg, getreg, arg);
}

int tweak_offset_set(unsigned value, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	return setregmask( mask(value,0,8), 0, 8, TWEAK_CRTC_ADDR, 0x13, setreg, getreg, arg);
}

int tweak_chained_set(int chained, int (*setreg)(unsigned value, unsigned port, int index, void* arg), unsigned (*getreg)(unsigned port, int index, void* arg), void* arg) {
	if (chained) {
		/* Turn on the Chain-4 bit */
		setregmask( 1, 3, 1, TWEAK_SEQ_ADDR, 0x4, setreg, getreg, arg);
		/* Turn on word mode (negate) */
		setregmask( 0, 6, 1, TWEAK_CRTC_ADDR, 0x17, setreg, getreg, arg);
		/* Turn on doubleword mode */
		setregmask( 1, 6, 1, TWEAK_CRTC_ADDR, 0x14, setreg, getreg, arg);
	} else {
		/* Turn off the Chain-4 bit */
		setregmask( 0, 3, 1, TWEAK_SEQ_ADDR, 0x4, setreg, getreg, arg);
		/* Turn off word mode (negate) */
		setregmask( 1, 6, 1, TWEAK_CRTC_ADDR, 0x17, setreg, getreg, arg);
		/* Turn off doubleword mode */
		setregmask( 0, 6, 1, TWEAK_CRTC_ADDR, 0x14, setreg, getreg, arg);
	}
	return 0;
}
